//  Copyright (c) 2001-2011 Hartmut Kaiser
// 
//  Distributed under the Boost Software License, Version 1.0. (See accompanying 
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#if !defined(BOOST_SPIRIT_LEXER_MARCH_22_2007_0929PM)
#define BOOST_SPIRIT_LEXER_MARCH_22_2007_0929PM

#if defined(_MSC_VER)
#pragma once
#endif

#include <boost/spirit/home/lex/lexer.hpp>
#include <boost/spirit/home/lex/qi.hpp>
#include <boost/spirit/home/lex/tokenize_and_parse.hpp>
#include <boost/spirit/home/lex/tokenize_and_parse_attr.hpp>

#endif
